"""Factory for platform-specific keyboard state implementations."""

import sys

from yek.platforms.base import KeyboardState
from yek.platforms.common import PynputKeyboardState
from yek.platforms.darwin import MacKeyboardState
from yek.platforms.linux import LinuxKeyboardState


def create_keyboard_state(max_buffer_len: int = 100) -> KeyboardState:
    """Return a keyboard state suitable for the current platform."""
    if sys.platform == "darwin":
        return MacKeyboardState(max_buffer_len)
    if sys.platform.startswith("linux"):
        return LinuxKeyboardState(max_buffer_len)

    # Fallback: use generic pynput-backed implementation
    return PynputKeyboardState(max_buffer_len)


__all__ = ["KeyboardState", "create_keyboard_state"]



"""Convenience constructors for common keyboard matcher keys."""

# pylint: disable=invalid-name,import-error

from contextlib import suppress

import pynput
import pynput.keyboard

from yek.matchers import Key, Matcher, _Or, StringMatcher

__all__ = [
    "Char",
    "String",
    "Alt",
    "AltL",
    "AltR",
    "AltGr",
    "Backspace",
    "CapsLock",
    "Cmd",
    "CmdL",
    "CmdR",
    "Ctrl",
    "CtrlL",
    "CtrlR",
    "Delete",
    "Down",
    "End",
    "Enter",
    "Esc",
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "F11",
    "F12",
    "F13",
    "F14",
    "F15",
    "F16",
    "F17",
    "F18",
    "F19",
    "F20",
    "Home",
    "Left",
    "PageDown",
    "PageUp",
    "Right",
    "Shift",
    "ShiftL",
    "ShiftR",
    "Space",
    "Tab",
    "Up",
    "MediaPlayPause",
    "MediaVolumeMute",
    "MediaVolumeDown",
    "MediaVolumeUp",
    "MediaPrevious",
]


def Char(char: str, case: bool = False) -> Matcher:
    """Create a matcher for a single character."""
    if not isinstance(char, str) and len(char) != 1:
        raise ValueError(f"We only support single character strings, got {char}")

    if case:
        return Key(name=char, code=repr(pynput.keyboard.KeyCode.from_char(char)))

    lower_case = Key(name=char, code=repr(pynput.keyboard.KeyCode.from_char(char.lower())))
    upper_case = Key(name=char, code=repr(pynput.keyboard.KeyCode.from_char(char.upper())))

    if lower_case != upper_case:
        return _Or(lower_case, upper_case)

    return lower_case


def String(s: str, case: bool = False) -> Matcher:
    """Create a matcher for an exact string sequence."""
    return StringMatcher(s, case=case)


Alt = Key(name="Alt", code=repr(pynput.keyboard.Key.alt.value))
AltL = Key(name="AltL", code=repr(pynput.keyboard.Key.alt_l.value))
AltR = Key(name="AltR", code=repr(pynput.keyboard.Key.alt_r.value))
AltGr = Key(name="AltGr", code=repr(pynput.keyboard.Key.alt_gr.value))
Backspace = Key(name="Backspace", code=repr(pynput.keyboard.Key.backspace.value))
CapsLock = Key(name="CapsLock", code=repr(pynput.keyboard.Key.caps_lock.value))
Cmd = Key(name="Cmd", code=repr(pynput.keyboard.Key.cmd.value))
CmdL = Key(name="CmdL", code=repr(pynput.keyboard.Key.cmd_l.value))
CmdR = Key(name="CmdR", code=repr(pynput.keyboard.Key.cmd_r.value))
Ctrl = Key(name="Ctrl", code=repr(pynput.keyboard.Key.ctrl.value))
CtrlL = Key(name="CtrlL", code=repr(pynput.keyboard.Key.ctrl_l.value))
CtrlR = Key(name="CtrlR", code=repr(pynput.keyboard.Key.ctrl_r.value))
Delete = Key(name="Delete", code=repr(pynput.keyboard.Key.delete.value))
Down = Key(name="Down", code=repr(pynput.keyboard.Key.down.value))
End = Key(name="End", code=repr(pynput.keyboard.Key.end.value))
Enter = Key(name="Enter", code=repr(pynput.keyboard.Key.enter.value))
Esc = Key(name="Esc", code=repr(pynput.keyboard.Key.esc.value))
F1 = Key(name="F1", code=repr(pynput.keyboard.Key.f1.value))
F2 = Key(name="F2", code=repr(pynput.keyboard.Key.f2.value))
F3 = Key(name="F3", code=repr(pynput.keyboard.Key.f3.value))
F4 = Key(name="F4", code=repr(pynput.keyboard.Key.f4.value))
F5 = Key(name="F5", code=repr(pynput.keyboard.Key.f5.value))
F6 = Key(name="F6", code=repr(pynput.keyboard.Key.f6.value))
F7 = Key(name="F7", code=repr(pynput.keyboard.Key.f7.value))
F8 = Key(name="F8", code=repr(pynput.keyboard.Key.f8.value))
F9 = Key(name="F9", code=repr(pynput.keyboard.Key.f9.value))
F10 = Key(name="F10", code=repr(pynput.keyboard.Key.f10.value))
F11 = Key(name="F11", code=repr(pynput.keyboard.Key.f11.value))
F12 = Key(name="F12", code=repr(pynput.keyboard.Key.f12.value))
F13 = Key(name="F13", code=repr(pynput.keyboard.Key.f13.value))
F14 = Key(name="F14", code=repr(pynput.keyboard.Key.f14.value))
F15 = Key(name="F15", code=repr(pynput.keyboard.Key.f15.value))
F16 = Key(name="F16", code=repr(pynput.keyboard.Key.f16.value))
F17 = Key(name="F17", code=repr(pynput.keyboard.Key.f17.value))
F18 = Key(name="F18", code=repr(pynput.keyboard.Key.f18.value))
F19 = Key(name="F19", code=repr(pynput.keyboard.Key.f19.value))
F20 = Key(name="F20", code=repr(pynput.keyboard.Key.f20.value))
Home = Key(name="Home", code=repr(pynput.keyboard.Key.home.value))
Left = Key(name="Left", code=repr(pynput.keyboard.Key.left.value))
PageDown = Key(name="PageDown", code=repr(pynput.keyboard.Key.page_down.value))
PageUp = Key(name="PageUp", code=repr(pynput.keyboard.Key.page_up.value))
Right = Key(name="Right", code=repr(pynput.keyboard.Key.right.value))
Shift = Key(name="Shift", code=repr(pynput.keyboard.Key.shift.value))
ShiftL = Key(name="ShiftL", code=repr(pynput.keyboard.Key.shift_l.value))
ShiftR = Key(name="ShiftR", code=repr(pynput.keyboard.Key.shift_r.value))
Space = Key(name="Space", code=repr(pynput.keyboard.Key.space.value))
Tab = Key(name="Tab", code=repr(pynput.keyboard.Key.tab.value))
Up = Key(name="Up", code=repr(pynput.keyboard.Key.up.value))
MediaPlayPause = Key(
    name="MediaPlayPause", code=repr(pynput.keyboard.Key.media_play_pause.value)
)
MediaVolumeMute = Key(
    name="MediaVolumeMute", code=repr(pynput.keyboard.Key.media_volume_mute.value)
)
MediaVolumeDown = Key(
    name="MediaVolumeDown", code=repr(pynput.keyboard.Key.media_volume_down.value)
)
MediaVolumeUp = Key(
    name="MediaVolumeUp", code=repr(pynput.keyboard.Key.media_volume_up.value)
)
MediaPrevious = Key(
    name="MediaPrevious", code=repr(pynput.keyboard.Key.media_previous.value)
)

with suppress(AttributeError):
    MediaNext = Key(name="MediaNext", code=repr(pynput.keyboard.Key.media_next.value))
    __all__.append("MediaNext")

with suppress(AttributeError):
    Insert = Key(name="Insert", code=repr(pynput.keyboard.Key.insert.value))
    __all__.append("Insert")

with suppress(AttributeError):
    Menu = Key(name="Menu", code=repr(pynput.keyboard.Key.menu.value))
    __all__.append("Menu")

with suppress(AttributeError):
    NumLock = Key(name="NumLock", code=repr(pynput.keyboard.Key.num_lock.value))
    __all__.append("NumLock")

with suppress(AttributeError):
    Pause = Key(name="Pause", code=repr(pynput.keyboard.Key.pause.value))
    __all__.append("Pause")

with suppress(AttributeError):
    PrintScreen = Key(
        name="PrintScreen", code=repr(pynput.keyboard.Key.print_screen.value)
    )
    __all__.append("PrintScreen")

with suppress(AttributeError):
    ScrollLock = Key(name="ScrollLock", code=repr(pynput.keyboard.Key.scroll_lock.value))
    __all__.append("ScrollLock")



"""Utilities for converting between matcher keys and pynput keys."""

from typing import Union

import pynput  # pylint: disable=import-error
import pynput.keyboard  # pylint: disable=import-error

from yek.events import KeyEvent, KeyEventKind
from yek.matchers import Key as MatcherKey

_KeyImpl = Union[str, pynput.keyboard.Key, pynput.keyboard.KeyCode, "KeyEvent", MatcherKey]


def get_key_code(k: _KeyImpl) -> Union[pynput.keyboard.KeyCode, str]:
    """Return a pynput KeyCode or string for the given input."""
    if isinstance(k, str):
        if len(k) != 1:
            raise ValueError(f"Only single character strings are supported, got {k}")
        return pynput.keyboard.KeyCode.from_char(k)

    if isinstance(k, pynput.keyboard.Key):
        return k.value

    if isinstance(k, pynput.keyboard.KeyCode):
        return k

    if isinstance(k, MatcherKey):
        return k.code

    raise ValueError(f"Unsupported key type {type(k)}")


def get_event(k: _KeyImpl, kind: KeyEventKind) -> KeyEvent:
    """Create a KeyEvent for the given key and kind."""
    return KeyEvent(key=get_key_code(k), kind=kind)



"""Shortcut orchestration and hot-reload utilities."""

# pylint: disable=missing-function-docstring,missing-class-docstring,too-few-public-methods
# pylint: disable=too-many-instance-attributes

import importlib
import importlib.util
import inspect
import os
import sys
import threading
import time
import uuid
from time import sleep
from typing import Any, Callable, Dict, List

from watchdog.events import FileSystemEventHandler, FileSystemEvent  # pylint: disable=import-error
from watchdog.observers import Observer  # pylint: disable=import-error

from yek.matchers import Matcher, Result
from yek.util import KeyEvent, KeyboardStateMap
from .action import Action, Context, SimpleAction

__all__ = ["App"]


def _debug(msg):
    if os.environ.get("DEBUG", False):
        print(msg)


_BUFFER_LIMIT = 5
_COMMAND_ATTR = "_is_command"
_ROUTE_ATTR = "_route"


class ActionTrigger:
    def __init__(self, matcher: Matcher, action: Action):
        self.action = action
        self.matcher = matcher
        self._last_match = 0

    def match(self, events: List[KeyEvent]) -> Result:
        events = [e for e in events if e.pressed_at > self._last_match]
        match = self.matcher.match(events)
        if match:
            self._last_match = time.time()
        return match


class ActionBankTrigger:
    def __init__(self):
        self._lock = threading.Lock()
        self._pairs: Dict[uuid.UUID, ActionTrigger] = {}
        self._properties = {}

    def match(self, events: List[KeyEvent]) -> List[Action]:
        with self._lock:
            actions = []
            for trigger in self._pairs.values():
                if trigger.match(events):
                    actions.append(trigger.action)
            return actions

    def register(self, trigger: ActionTrigger) -> None:
        with self._lock:
            self._pairs[trigger.action.id] = trigger

    def find_by_prop(self, prop: str, value: str) -> List[ActionTrigger]:
        registers = []
        for _, register in self._pairs.items():
            prop_value = register.action.properties.get(prop)
            if prop_value == value:
                registers.append(register)
        return registers

    def remove_by_prop(self, prop: str, value: Any):
        with self._lock:
            for register in self.find_by_prop(prop, value):
                self._pairs.pop(register.action.id, None)


def _load_action_triggers(file_path: str) -> List[ActionTrigger]:
    try:
        spec = importlib.util.spec_from_file_location("module", file_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules["module"] = module
        spec.loader.exec_module(module)

        route_mapping = []
        app_instance = None

        # First, find the App instance
        for _, obj in inspect.getmembers(module):
            if isinstance(obj, App):
                app_instance = obj
                break

        if app_instance is None:
            return []

        # Look for route decorators
        for _, obj in inspect.getmembers(module):
            if isinstance(obj, ActionTrigger):
                route_mapping.append(obj)
        return route_mapping
    except SyntaxError:
        return []
    finally:
        # Clean up the temporary module from sys.modules
        if "module" in sys.modules:
            del sys.modules["module"]


class _HotReloader(FileSystemEventHandler):
    def __init__(self, app: "App"):
        self.app = app
        self.last_modified = 0

    def on_modified(self, event: FileSystemEvent):
        if self.app.action_triggers.find_by_prop("file", event.src_path):
            current_time = time.time()
            if current_time - self.last_modified > 1:  # Debounce
                self.last_modified = current_time
                _debug(f"Detected change in {event.src_path}. Reloading routes...")
                self.reload_routes(event.src_path)

    def reload_routes(self, path: str):
        try:
            triggers = _load_action_triggers(path)
            self.app.action_triggers.remove_by_prop("file", path)
            for trigger in triggers:
                print(f"Registering: {trigger.action} as {trigger.matcher.debug()}")
                self.app.action_triggers.register(trigger)
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"Failed to reload routes: {e}")  # pragma: no cover
        else:
            print("Routes reloaded successfully.")


class App:
    def __init__(self,
                 buffer_timeout: float = _BUFFER_LIMIT,
                 sequence_timeout: float = 2,
                 ):
        self.action_triggers: ActionBankTrigger = ActionBankTrigger()
        self._buffer_timeout = buffer_timeout
        self._sequence_timeout = sequence_timeout
        self._running = threading.Event()
        self._lock = threading.Lock()
        self._last_called: float = 0

        self._keyboard = KeyboardStateMap()  # pylint: disable=abstract-class-instantiated

        # Hot-reload
        self._observer = None
        self._reloader = None
        self.routes_file = None

        Matcher.app = self

    def on(self, matcher: Matcher) -> Callable:
        _debug(matcher.debug())

        def _decorator(func: Callable) -> ActionTrigger:
            if not isinstance(func, Callable):
                raise ValueError("on decorator must be used with a callable")

            trigger = ActionTrigger(matcher, SimpleAction(func))
            self.action_triggers.register(trigger)
            return trigger

        return _decorator

    def __call__(self):
        self._keyboard.start()
        self._running.set()

        # Hot-reload
        self._setup_hot_reload()

        try:
            while self._running.is_set():
                # self._process_running_loops()
                self._check_shortcuts()
                sleep(0.01)
        except KeyboardInterrupt:
            self.stop()

    def stop(self):
        self._keyboard.stop()
        self._running.clear()
        if self._observer:
            self._observer.stop()
            self._observer.join()

    def _since(self):
        window = min(self._buffer_timeout, self._sequence_timeout)
        return max(self._last_called, time.time() - window)

    def _check_shortcuts(self):
        keys = self._keyboard.get_since(self._since())
        if len(keys) == 0:
            return

        actions = self.action_triggers.match(keys)
        for action in actions:
            action.execute(Context(self._keyboard))
        if actions:
            self._last_called = time.time()

    def _setup_hot_reload(self):
        # Get the calling module (where the routes are defined)
        calling_frame = inspect.stack()[2]
        module = inspect.getmodule(calling_frame[0])
        self.routes_file = module.__file__

        self._reloader = _HotReloader(self)
        self._observer = Observer()
        self._observer.schedule(
            self._reloader, path=os.path.dirname(self.routes_file), recursive=False
        )
        self._observer.start()



"""Platform-specific keyboard state abstraction."""

# pylint: disable=missing-function-docstring

import abc
from typing import List

from yek.events import KeyEvent
from yek.matchers import Key


class KeyboardState(abc.ABC):
    """Interface implemented by each platform backend."""

    @abc.abstractmethod
    def start(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def is_pressed(self, key: Key) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def is_pressed_any(self, *keys: Key) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def is_only_pressed(self, *keys: Key) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def was_pressed_any_since(self, since: float, *keys: Key) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def other_than_was_pressed_since(self, since: float, *keys: Key) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def get_since(self, since: float) -> List["KeyEvent"]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_in_the_last(self, seconds: float) -> List["KeyEvent"]:
        raise NotImplementedError



"""Time utilities for matcher sequencing."""

# pylint: disable=missing-function-docstring


class Wait:
    """Wrapper for wait durations used by matchers."""

    def __init__(self, milli: int = 0, seconds: int = 0):
        if milli < 0 or seconds < 0:
            raise ValueError("Time cannot be negative")

        if milli == 0 and seconds == 0:
            raise ValueError("Time cannot be zero")

        self._milliseconds = milli
        self._seconds = seconds

    @property
    def milliseconds(self):
        return self._milliseconds + self._seconds * 1000

    @property
    def seconds(self):
        return self._seconds + self._milliseconds / 1000

    def __float__(self) -> float:
        return self.seconds



"""Thread-safe containers used by the shortcut engine."""

# pylint: disable=missing-function-docstring

import threading
import time
from abc import ABC, abstractmethod
from collections import OrderedDict, deque
from typing import Deque, Generic, List, TypeVar


class LimitedMap(ABC):
    """Map with eviction policy implemented by subclasses."""

    def __init__(self):
        self.data = OrderedDict()
        self.lock = threading.Lock()

    @abstractmethod
    def _evict(self):
        pass

    def __setitem__(self, key, value):
        with self.lock:
            if key in self.data:
                del self.data[key]
            self.data[key] = (value, time.time())
            self._evict()

    def __getitem__(self, key):
        with self.lock:
            if key in self.data:
                value, _ = self.data[key]
                return value
            raise KeyError(key)

    def __delitem__(self, key):
        with self.lock:
            del self.data[key]

    def __contains__(self, key):
        with self.lock:
            return key in self.data

    def __len__(self):
        with self.lock:
            return len(self.data)

    def keys(self):
        with self.lock:
            return list(self.data.keys())

    def values(self):
        with self.lock:
            return [value for value, _ in self.data.values()]

    def items(self):
        with self.lock:
            return [(key, value) for key, (value, _) in self.data.items()]

    def get(self, key, default=None):
        with self.lock:
            try:
                return self[key]
            except KeyError:
                return default

    def clear(self):
        with self.lock:
            self.data.clear()


class CountLimitedMap(LimitedMap):
    """Evicts items after a maximum count is reached."""

    def __init__(self, limit):
        super().__init__()
        self.limit = limit

    def _evict(self):
        while len(self.data) > self.limit:
            self.data.popitem(last=False)


class TimeLimitedMap(LimitedMap):
    """Evicts items older than the configured age (in seconds)."""

    def __init__(self, limit):
        super().__init__()
        self.limit = limit

    def _evict(self):
        current_time = time.time()
        keys_to_remove = [
            key
            for key, (_, timestamp) in self.data.items()
            if current_time - timestamp > self.limit
        ]
        for key in keys_to_remove:
            del self.data[key]


T = TypeVar("T")


class ThreadSafeBuffer(Generic[T]):
    """Fixed-size buffer that can be read safely from multiple threads."""

    def __init__(self, max_len: int):
        self._buffer: Deque[T] = deque(maxlen=max_len)
        self.lock = threading.Lock()
        self.last_added = 0

    def append(self, key: T):
        with self.lock:
            self._buffer.append(key)

    def get(self) -> List[T]:
        with self.lock:
            return list(self._buffer)

    def get_since(self, since: float) -> List[T]:
        added = []
        with self.lock:
            for item in self._buffer:
                if item.pressed_at >= since:
                    added.append(item)
        return added



"""Actions and execution context for shortcuts."""

# pylint: disable=missing-class-docstring,missing-function-docstring

import abc
import inspect
import threading
import time
import uuid
from typing import Any, Callable, Dict, List, Optional, Union

from pynput.keyboard import Key, KeyCode  # pylint: disable=import-error

from yek.util import KeyboardStateMap, KeyEvent, get_function_details


class Action(abc.ABC):
    @abc.abstractmethod
    def execute(self, context: "Context"):
        pass

    @abc.abstractmethod
    def is_running(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    def __eq__(self, other: "Action"):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __str__(self) -> str:
        return self.id

    @property
    @abc.abstractmethod
    def properties(self) -> Optional[Any]:
        pass


class SimpleAction(Action):
    def __init__(self, func: Callable[..., None]):
        self._id = str(uuid.uuid4())
        desc = get_function_details(func)
        self._props = {
            "file": desc.file,
            "name": desc.name,
            "line": desc.line
        }
        self._func = func
        self._execution = None
        self._lock = threading.Lock()

    def execute(self, context: "Context"):
        with self._lock:
            if self.is_running():
                return
            self._execution = threading.Thread(target=self._func, args=(context,))
            self._execution.start()

    def is_running(self) -> bool:
        if self._execution is None:
            return False
        return self._execution.is_alive()

    @property
    def id(self) -> str:
        return self._id

    @property
    def properties(self) -> Optional[Any]:
        return self._props

    def __str__(self) -> str:
        name, file, line = self._props["name"], self._props["file"], self._props["line"]
        return f"SimpleAction(func={name}, defined_at={file}:{line})"


class Context:
    def __init__(self, keyboard_state: KeyboardStateMap):
        self.__keyboard_state = keyboard_state
        self.__started_at = time.time()
        self.__payload = {}

    def is_pressed(self, key: Key):
        return self.__keyboard_state.is_pressed(key)

    def is_only_pressed(self, *keys: Union[KeyCode, KeyEvent, Key]):
        return self.__keyboard_state.is_only_pressed(*keys)

    def was_pressed_other_than(self, *keys: Union[KeyCode, KeyEvent, Key]):
        return self.__keyboard_state.other_than_was_pressed_since(
            self.__started_at, *keys
        )

    def keys(self) -> List[KeyEvent]:
        return self.__keyboard_state.get_since(self.__started_at)

    @property
    def started_at(self):
        return self.__started_at

    @property
    def payload(self) -> Dict[str, Any]:
        return self.__payload


def get_function_name(obj: Callable[..., Any]) -> Optional[str]:
    # Check if the object is a function or method
    if inspect.isfunction(obj) or inspect.ismethod(obj):
        return obj.__name__
    return None



"""Utility helpers for keyboard state and function metadata."""

import inspect
import os
from dataclasses import dataclass
from typing import Any, Callable

from yek.events import KeyEvent
from yek.platforms import KeyboardState, create_keyboard_state

__all__ = ["KeyboardStateMap", "KeyEvent"]

_DEBUG = os.environ.get("DEBUG", False)


def _debug(msg):
    if _DEBUG:
        print(msg)


class KeyboardStateMap(KeyboardState):
    """
    Factory wrapper returning a platform-specific keyboard state implementation.
    """

    def __new__(cls, max_buffer_len: int = 100):  # type: ignore[override]
        return create_keyboard_state(max_buffer_len)


@dataclass
class FunctionDetails:
    """Captured metadata about a callable."""
    name: str
    line: int
    file: str


def get_function_details(obj: Callable[..., Any]) -> FunctionDetails:
    """Extract function name, line number, and file path."""
    # Check if the object is a function or method
    if inspect.isfunction(obj) or inspect.ismethod(obj):
        name = obj.__name__
        line_number = inspect.getsourcelines(obj)[1]
        file_name = inspect.getfile(obj)
        return FunctionDetails(name=name, line=line_number, file=file_name)

    raise ValueError(f"Unsupported object type {type(obj)} - expected function or method")



"""Keyboard event structures."""

# pylint: disable=missing-function-docstring,import-error

import time
import uuid
from enum import Enum
from typing import Optional, Sequence, Union

import pynput


class KeyEventKind(Enum):
    """Kinds of key events we track."""

    PRESSED = "PRESSED"
    RELEASED = "RELEASED"


class KeyEvent:
    """Key event with timestamp and unique identifier."""

    def __init__(self, key: Union[pynput.keyboard.KeyCode, str], kind: KeyEventKind):
        self.id = uuid.uuid4()
        self.key = key
        self.pressed_at = time.time()
        self.kind = kind

    def __eq__(self, other: "KeyEvent"):
        if not isinstance(other, KeyEvent):
            return False

        return (
            self.code == other.code
            and self.kind == other.kind
        )

    @property
    def code(self) -> str:
        return self.key if isinstance(self.key, str) else repr(self.key)

    def get_key_next_event(self, events: Sequence["KeyEvent"]) -> Optional["KeyEvent"]:
        for event in events:
            if event.pressed_at > self.pressed_at and event.key == self.key:
                return event
        return None

    def __repr__(self):
        return f"KeyEvent({self.key} - {self.kind} - {self.pressed_at})"

    def __hash__(self):
        return hash((self.code, self.kind))

    def __str__(self):
        return str(self.key)



"""Keyboard matchers and combinators."""

# pylint: disable=missing-class-docstring,missing-function-docstring

import abc
import math
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional, Sequence, Tuple, Union

from yek.events import KeyEvent, KeyEventKind
from yek.time import Wait

if TYPE_CHECKING:
    from yek import App

__all__ = [
    "Key",
    "StringMatcher",
    "Matcher",
    "_TimedPressRelease",
    "_MatchSequence",
    "Hold",
    "Throttle",
]


@dataclass
class Match:
    start: KeyEvent
    end: KeyEvent


@dataclass
class Result:
    stop_position: int
    value: Optional[Match] = None

    def __bool__(self) -> bool:
        return self.value is not None


class Matcher(abc.ABC):
    app: "App" = None

    @abc.abstractmethod
    def match(self, events: Sequence[KeyEvent]) -> Result:
        """Return a Result when events satisfy the matcher."""

    @abc.abstractmethod
    def __truediv__(self, other: Union["Matcher", Tuple[int, int]]) -> "Matcher":
        """Compose matchers in sequence."""

    @abc.abstractmethod
    def __or__(self, other: "Matcher") -> "Matcher":
        """Alternate between two matchers."""

    @abc.abstractmethod
    def __and__(self, other: "Matcher") -> "Matcher":
        """Require both matchers."""

    @abc.abstractmethod
    def __matmul__(self, other: Union[float, Wait]) -> "Matcher":
        """Combine with a timing constraint."""

    @abc.abstractmethod
    def debug(self) -> str:
        """Debug description of the matcher."""

    def throttle(self, every_ms: int = 100) -> "Throttle":
        """Wrap this matcher in a Throttle."""
        return Throttle(self, every_ms=every_ms)


class Key(Matcher):
    def __init__(
            self,
            name: str,
            code: str,
            char: Optional[str] = None,
            kind: KeyEventKind = KeyEventKind.PRESSED,
    ):
        self._name = name
        self._code = code
        self._char = char
        self._event_kind = kind

    def __hash__(self) -> int:
        return hash(self._code)

    def __repr__(self) -> str:
        return f"Key({self._name})"

    def debug(self) -> str:
        return f"Key({self._name})"

    @property
    def code(self) -> str:
        return self._code

    def match(self, events: Sequence[KeyEvent]) -> Result:
        for n, event in enumerate(events):
            if event.code == self._code and event.kind == self._event_kind:
                return Result(value=Match(start=event, end=event), stop_position=n)
        return Result(stop_position=len(events))

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        def _to_seconds(value: Union[float, int, Wait]) -> float:
            if isinstance(value, Wait):
                return value.seconds
            if isinstance(value, (int, float)):
                return float(value)
            raise ValueError(f"Unsupported wait value type {type(value)}")

        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError("Timed press-release tuple must have exactly 2 values")
            start, end = (_to_seconds(other[0]), _to_seconds(other[1]))
            if not 0 <= start < end:
                raise ValueError("Timed press-release must satisfy 0 <= start < end")
            return _TimedPressRelease(self, start=start, end=end)

        duration = _to_seconds(other)
        if duration <= 0:
            raise ValueError("Timed press-release duration must be positive")
        return _TimedPressRelease(self, start=0, end=duration)

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> Matcher:
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)


class StringMatcher(Matcher):
    def __init__(self, string: str, case: bool = True):
        self._string = string
        self._case = case

    def __hash__(self) -> int:
        return hash(self._string)

    def __repr__(self) -> str:
        return f"String({self._string})"

    def debug(self) -> str:
        return f"String({self._string})"

    def match(self, events: Sequence[KeyEvent]) -> Result:
        matched_indexes = []
        i = 0
        for char in self._string:
            while i < len(events):
                # Hacky way to ignore non-character keys
                event_char = str(events[i].key).strip("'")
                if len(char) > 1:
                    i += 1
                    continue

                event_char = event_char if self._case else event_char.lower()
                normalized = char if self._case else char.lower()

                if event_char == normalized:
                    matched_indexes.append(i)
                    i += 1
                    break

                i += 1
            else:
                return Result(stop_position=i)

        return Result(
            value=Match(
                start=events[matched_indexes[0]],
                end=events[matched_indexes[-1]]
            ),
            stop_position=i
        )

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        raise ValueError("Cannot chain timed press-release with a string matcher")

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> Matcher:
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)


class _TimedPressRelease(Matcher):
    def __init__(self, key: "Key", start: float, end: float):
        self._key = key
        self._start = start
        self._end = end

    def match(self, events: Sequence[KeyEvent]) -> Result:
        for i, event in enumerate(events):
            if event.code == self._key.code and event.kind == KeyEventKind.PRESSED:
                for j, event2 in enumerate(events[i + 1:]):
                    if event2.code == self._key.code and event2.kind == KeyEventKind.RELEASED:
                        if self._start <= event2.pressed_at - event.pressed_at <= self._end:
                            return Result(value=Match(start=event, end=event2), stop_position=i + j)
        return Result(stop_position=len(events))

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        raise ValueError("Cannot chain timed press-release with another timed press-release")

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"TimedPressRelease({self._key.debug()}, {self._start}, {self._end})"


class _MatchSequence(Matcher):
    def __init__(self, *matchers: Matcher):
        self._matchers = []
        for matcher in matchers:
            if isinstance(matcher, _MatchSequence):
                self._matchers.extend(matcher._matchers)
            else:
                self._matchers.append(matcher)
        self._hold_tail = _is_hold_like(self._matchers[-1]) if self._matchers else False
        self._armed = False

    def match(self, events: Sequence[KeyEvent]) -> Result:
        if self._hold_tail and self._armed:
            tail = self._matchers[-1]
            tail_result = tail.match(events)
            if tail_result:
                return tail_result
            self._armed = False
            return Result(stop_position=len(events))

        position = 0
        matches = []
        for matcher in self._matchers:
            result = matcher.match(events[position:])
            if not result:
                return Result(stop_position=len(events))

            matches.append(result.value)
            position += result.stop_position + 1

        stop_position = max(position - 1, 0)
        if self._hold_tail:
            self._armed = True
        return Result(
            value=Match(start=matches[0].start, end=matches[-1].end),
            stop_position=stop_position,
        )

    def __matmul__(self, other) -> "Matcher":
        raise ValueError("Cannot chain timed press-release with a match sequence")

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"MatchSequence({', '.join([m.debug() for m in self._matchers])})"


class _Or(Matcher):
    def __init__(self, *matchers: Matcher):
        self._matchers = matchers

    def match(self, events: Sequence[KeyEvent]) -> Result:
        for matcher in self._matchers:
            match = matcher.match(events)
            if match:
                return match
        return Result(stop_position=len(events))

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        return _Or(*[m.__matmul__(other) for m in self._matchers])

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"Or({', '.join([m.debug() for m in self._matchers])})"


class _And(Matcher):
    def __init__(self, *matchers: Matcher):
        self._matchers = matchers

    def match(self, events: Sequence[KeyEvent]) -> Result:
        matches = []
        stop_position = 0
        for matcher in self._matchers:
            match = matcher.match(events)
            if not match:
                return Result(stop_position=len(events))

            stop_position = max(stop_position, match.stop_position)
            matches.append(match.value)
        return Result(
            value=Match(start=matches[0].start, end=matches[-1].end),
            stop_position=stop_position,
        )

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        return _and(*[m.__matmul__(other) for m in self._matchers])

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"And({', '.join([m.debug() for m in self._matchers])})"


class _AndKeys(Matcher):
    def __init__(self, *keys: Key):
        for k in keys:
            if not isinstance(k, Key):
                raise ValueError(f"Expected Key instance, got {type(k)}")

        self._keys = keys

    def match(self, events: Sequence[KeyEvent]) -> Result:
        intervals = []
        min_stop, max_stop = math.inf, 0
        for key in self._keys:
            match = key.match(events)
            min_stop = min(min_stop, match.stop_position)
            max_stop = max(max_stop, match.stop_position)

            if not match:
                return Result(stop_position=len(events))

            press_event = match.value.start
            release_event = press_event.get_key_next_event(events)

            intervals.append((press_event, release_event))

        intersection = get_common_intersection(intervals)

        if intersection:
            return Result(value=Match(
                start=intersection[0],
                end=intersection[1]
            ), stop_position=max_stop)

        return Result(stop_position=min_stop)

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        return _and(*[k.__matmul__(other) for k in self._keys])

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> Matcher:
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"AndKeys({', '.join([k.debug() for k in self._keys])})"


class _AndStartEndMatcher(Matcher):
    def __init__(self, a: Key, b: Union[_TimedPressRelease, _MatchSequence, StringMatcher]):
        self.a = a
        self.b = b

    def match(self, events: Sequence[KeyEvent]) -> Result:
        match_a = self.a.match(events)
        if not match_a:
            return Result(stop_position=match_a.stop_position)
        end_a = match_a.value.start.get_key_next_event(events)

        match_b = self.b.match(events)
        if not match_b:
            return Result(stop_position=match_b.stop_position)

        intersection = get_common_intersection([
            (match_a.value.start, end_a),
            (match_b.value.start, match_b.value.end)
        ])

        if intersection:
            stop = max(match_a.stop_position, match_b.stop_position)
            return Result(
                value=Match(start=intersection[0], end=intersection[1]),
                stop_position=stop,
            )

        stop = min(match_a.stop_position, match_b.stop_position)
        return Result(stop_position=stop)

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        return _and(self.a.__matmul__(other), self.b.__matmul__(other))

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> Matcher:
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def debug(self) -> str:
        return f"AndStartEndMatcher({self.a.debug()}, {self.b.debug()})"


def get_common_intersection(
    events: List[Tuple[Optional[KeyEvent], Optional[KeyEvent]]]
) -> Optional[Tuple[Optional[KeyEvent], Optional[KeyEvent]]]:
    if not events:
        return None

    cache = {}
    # Sort events by start time, treating None as negative infinity
    sorted_events = sorted(events, key=lambda x: x[0].pressed_at if x[0] else 0)

    # Initialize the intersection
    intersection_start = sorted_events[0][0].pressed_at if sorted_events[0][0] else 0
    intersection_end = sorted_events[0][1].pressed_at if sorted_events[0][1] else math.inf

    cache[intersection_start] = sorted_events[0][0]
    cache[intersection_end] = sorted_events[0][1]

    # Iterate through the rest of the events
    for start, end in sorted_events[1:]:
        start_time = start.pressed_at if start else 0
        end_time = end.pressed_at if end else math.inf

        cache[start_time] = start
        cache[end_time] = end

        # If the current event starts after the intersection ends, there's no overlap
        if start_time >= intersection_end:
            return None

        # Update the intersection
        intersection_start = max(intersection_start, start_time)
        intersection_end = min(intersection_end, end_time)

    # Return the final intersection interval
    return cache.get(intersection_start), cache.get(intersection_end)


def _and(*a: Matcher) -> Matcher:
    # pylint: disable=too-many-return-statements,too-many-branches,protected-access
    if len(a) == 0:
        raise ValueError("Cannot AND zero matchers")
    if len(a) == 1:
        return a[0]
    if len(a) > 2:
        return _and(a[0], _and(a[1], *a[2:]))

    a, b = a
    if isinstance(a, Key) and isinstance(b, Key):
        return _AndKeys(a, b)

    if isinstance(a, _AndKeys) and isinstance(b, Key):
        return _AndKeys(*a._keys, b)

    if isinstance(a, Key) and isinstance(b, _AndKeys):
        return _AndKeys(*b._keys, a)

    if isinstance(a, Key) and isinstance(b, (_TimedPressRelease, _MatchSequence, StringMatcher)):
        return _AndStartEndMatcher(a, b)

    if isinstance(a, (_TimedPressRelease, _MatchSequence, StringMatcher)) and isinstance(b, Key):
        return _AndStartEndMatcher(a=b, b=a)

    if isinstance(a, _Or) and isinstance(b, (_TimedPressRelease, _MatchSequence, StringMatcher)):
        return _Or(*[_and(m, b) for m in a._matchers])

    if isinstance(a, (_TimedPressRelease, _MatchSequence, StringMatcher)) and isinstance(b, _Or):
        return _Or(*[_and(a, m) for m in b._matchers])

    if isinstance(a, _Or) and isinstance(b, _Or):
        return _Or(*[_and(m, b) for m in a._matchers])

    if isinstance(a, Key) and isinstance(b, _Or):
        return _Or(*[_and(m, a) for m in b._matchers])

    if isinstance(a, _AndKeys) and isinstance(b, _Or):
        return _Or(*[_and(m, a) for m in b._matchers])

    if isinstance(a, _Or) and isinstance(b, _AndKeys):
        return _Or(*[_and(m, b) for m in a._matchers])

    return _And(a, b)


def _is_hold_like(matcher: Matcher) -> bool:
    if isinstance(matcher, Hold):
        return True
    if isinstance(matcher, Throttle):
        return _is_hold_like(matcher.inner_matcher)
    return False


class Hold(Matcher):
    """
    Matches when the given keys are currently held down (live keyboard state),
    without consuming any events from the sequence.
    """

    def __init__(self, *keys: Key, only: bool = False):
        if len(keys) == 0:
            raise ValueError("Hold requires at least one key")
        for key in keys:
            if not isinstance(key, Key):
                raise ValueError(f"Hold expects Key instances, got {type(key)}")
        self._keys = keys
        self._only = only

    def match(self, events: Sequence[KeyEvent]) -> Result:
        keyboard_state = getattr(self.app, "_keyboard", None) if self.app else None
        if keyboard_state is None:
            return Result(stop_position=len(events))

        if self._only:
            if not keyboard_state.is_only_pressed(*self._keys):
                return Result(stop_position=len(events))
        elif not all(keyboard_state.is_pressed(k) for k in self._keys):
            return Result(stop_position=len(events))

        synthetic = KeyEvent(key="__hold__", kind=KeyEventKind.PRESSED)
        synthetic.pressed_at = time.time()
        return Result(value=Match(start=synthetic, end=synthetic), stop_position=-1)

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        raise ValueError("Cannot chain timed press-release with Hold")

    def debug(self) -> str:
        suffix = ", only=True" if self._only else ""
        return f"Hold({', '.join([k.debug() for k in self._keys])}{suffix})"

    def throttle(self, every_ms: int = 100) -> "Throttle":
        """Convenience to wrap this hold in a Throttle."""
        return Throttle(self, every_ms=every_ms)


class Throttle(Matcher):
    """
    Wraps another matcher and limits how often it can return a match.
    Useful for held combos that should only fire every N milliseconds.
    """

    def __init__(self, matcher: Matcher, every_ms: int = 100):
        if every_ms <= 0:
            raise ValueError("every_ms must be positive")
        self._matcher = matcher
        self._every_ms = every_ms
        self._last_fire = 0.0
        self.inner_matcher = matcher  # public alias for composed checks

    def match(self, events: Sequence[KeyEvent]) -> Result:
        result = self._matcher.match(events)
        if not result:
            return result

        now = time.time()
        if now - self._last_fire >= self._every_ms / 1000:
            self._last_fire = now
            return result

        return Result(stop_position=result.stop_position)

    def __truediv__(self, other: Union[Matcher, Tuple[int, int]]) -> "Matcher":
        return _MatchSequence(self, other)

    def __or__(self, other: Matcher) -> Matcher:
        return _Or(self, other)

    def __and__(self, other: Matcher) -> Matcher:
        return _and(self, other)

    def __matmul__(self, other: Union[float, Wait]) -> Matcher:
        raise ValueError("Cannot chain timed press-release with Throttle")

    def debug(self) -> str:
        return f"Throttle({self._matcher.debug()}, every_ms={self._every_ms})"



"""macOS-specific keyboard state helpers."""

import os
import subprocess
from contextlib import suppress

from yek.platforms.common import PynputKeyboardState


class MacKeyboardState(PynputKeyboardState):
    """
    macOS keyboard state that nudges the user to grant Input Monitoring permission.
    """

    _prompted = False

    def start(self):
        try:
            super().start()
        except Exception as exc:
            self._prompt_permissions(exc)
            raise

    def _prompt_permissions(self, exc: Exception):
        del exc
        if MacKeyboardState._prompted:
            return
        if os.environ.get("YEK_SKIP_MAC_PROMPT"):
            return

        MacKeyboardState._prompted = True
        with suppress(Exception):
            print(
                "yek could not start keyboard listener (macOS permissions may be missing). "
                "Opening Input Monitoring settings..."
            )
            subprocess.Popen(  # pylint: disable=consider-using-with
                [
                    "open",
                    "x-apple.systempreferences:com.apple.preference.security?Privacy_Keyboard",
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )



"""Shared pynput-backed keyboard state implementation."""

# pylint: disable=missing-function-docstring,import-error

import threading
import time
from typing import Dict, List, Union

import pynput
import pynput.keyboard

from yek.data_structures import ThreadSafeBuffer
from yek.events import KeyEvent, KeyEventKind
from yek.key_utils import get_event
from yek.platforms.base import KeyboardState


class PynputKeyboardState(KeyboardState):
    """Keyboard state backed by pynput listeners."""

    def __init__(self, max_buffer_len: int = 100):
        self._running = threading.Event()
        self._lock = threading.Lock()
        self._listener = None
        self._map: Dict[str, bool] = {}
        self._buffer = ThreadSafeBuffer[KeyEvent](max_len=max_buffer_len)

    def start(self):
        self._running.set()
        self._listener = pynput.keyboard.Listener(
            on_press=self._on_press,
            on_release=self._on_release
        )
        self._listener.start()

    def stop(self):
        self._running.clear()
        if self._listener:
            self._listener.stop()

    def _on_press(self, key: Union[pynput.keyboard.Key, pynput.keyboard.KeyCode]):
        event = get_event(key, kind=KeyEventKind.PRESSED)
        with self._lock:
            self._buffer.append(event)
            self._map[event.code] = True

    def _on_release(self, key: Union[pynput.keyboard.Key, pynput.keyboard.KeyCode]):
        event = get_event(key, kind=KeyEventKind.RELEASED)
        with self._lock:
            self._buffer.append(event)
            self._map[event.code] = False

    def is_pressed(self, key) -> bool:
        with self._lock:
            return self._map.get(key.code, False)

    def is_pressed_any(self, *keys) -> bool:
        codes = {k.code for k in keys}
        with self._lock:
            return any(self._map.get(code, False) for code in codes)

    def is_only_pressed(self, *keys) -> bool:
        keys = {k.code for k in keys}
        with self._lock:
            return keys == {k for k, v in self._map.items() if v}

    def was_pressed_any_since(self, since: float, *keys) -> bool:
        keys = {get_event(k, kind=KeyEventKind.PRESSED) for k in keys}
        buffer = set(self._buffer.get_since(since))
        return bool(buffer.intersection(keys))

    def other_than_was_pressed_since(self, since: float, *keys) -> bool:
        keys = {get_event(k, kind=KeyEventKind.PRESSED) for k in keys}
        buffer = set(self._buffer.get_since(since))
        return bool(buffer.difference(keys))

    def get_since(self, since: float) -> List["KeyEvent"]:
        return self._buffer.get_since(since)

    def get_in_the_last(self, seconds: float) -> List["KeyEvent"]:
        return self._buffer.get_since(time.time() - seconds)
